from __future__ import annotations

from contextlib import closing
import os
from pathlib import Path
import re
import secrets
from typing import Iterator, Sequence

import pexpect
import pytest

from ai.backend.test.utils.cli import ClientRunnerFunc, EOF, run as _run

_rx_env_export = re.compile(r"^(export )?(?P<key>\w+)=(?P<val>.*)$")


@pytest.fixture(scope="session")
def client_venv() -> Path:
    p = os.environ.get("BACKENDAI_TEST_CLIENT_VENV", None)
    if p is None:
        raise RuntimeError("Missing BACKENDAI_TEST_CLIENT_VENV env-var!")
    return Path(p)


@pytest.fixture(scope="session")
def client_bin(
    client_venv: Path,
) -> Path:
    return client_venv / 'bin' / 'backend.ai'


@pytest.fixture(scope="session")
def client_environ() -> dict[str, str]:
    p = os.environ.get("BACKENDAI_TEST_CLIENT_ENV", None)
    if p is None:
        raise RuntimeError("Missing BACKENDAI_TEST_CLIENT_ENV env-var!")
    envs = {}
    sample_admin_sh = Path(p)
    if sample_admin_sh.exists():
        lines = sample_admin_sh.read_text().splitlines()
        for line in lines:
            if m := _rx_env_export.search(line.strip()):
                envs[m.group('key')] = m.group('val')
    return envs


@pytest.fixture(scope="session")
def run(client_bin: Path, client_environ: dict[str, str]) -> Iterator[ClientRunnerFunc]:

    def run_impl(cmdargs: Sequence[str | Path], *args, **kwargs) -> pexpect.spawn:
        return _run([client_bin, *cmdargs], *args, **kwargs, env=client_environ)

    yield run_impl


@pytest.fixture
def domain_name() -> str:
    return f"testing-{secrets.token_hex(8)}"


@pytest.fixture
def temp_domain(domain_name: str, run: ClientRunnerFunc) -> Iterator[str]:
    run(['admin', 'domains', 'add', domain_name])
    print("==== temp_domain created ====")
    try:
        yield domain_name
    finally:
        with closing(run(['admin', 'domains', 'purge', domain_name])) as p:
            p.expect_exact("Are you sure?")
            p.sendline("Y")
            p.expect(EOF)



import json
from contextlib import closing

from ...utils.cli import EOF, ClientRunnerFunc


def test_add_user(run: ClientRunnerFunc):
    """
    Testcase for user addition.
    """
    print("[ Add user ]")

    # Check if test account exists
    with closing(run(['--output=json', 'admin', 'user', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        user_list = loaded.get('items')

    test_user1 = get_user_from_list(user_list, 'testaccount1')
    test_user2 = get_user_from_list(user_list, 'testaccount2')
    test_user3 = get_user_from_list(user_list, 'testaccount3')

    if not bool(test_user1):
        # Add user
        add_arguments = [
            '--output=json', 'admin', 'user', 'add',
            '-u', 'testaccount1',
            '-n', 'John Doe',
            '--need-password-change', 'default',
            'testaccount1@lablup.com',
            '1q2w3e4r',
        ]
        with closing(run(add_arguments)) as p:
            p.expect(EOF)
            assert 'User testaccount1@lablup.com is created' in p.before.decode(), 'Account add error'

    if not bool(test_user2):
        # Add user
        add_arguments = [
            '--output=json', 'admin', 'user', 'add',
            '-u', 'testaccount2',
            '-n', 'John Roe',
            '-r', 'admin',
            '-s', 'inactive',
            'default',
            'testaccount2@lablup.com',
            '1q2w3e4r',
        ]
        with closing(run(add_arguments)) as p:
            p.expect(EOF)
            assert 'User testaccount2@lablup.com is created' in p.before.decode(), 'Account add error'

    if not bool(test_user3):
        # Add user
        add_arguments = [
            '--output=json', 'admin', 'user', 'add',
            '-u', 'testaccount3',
            '-n', 'Richard Roe',
            '-r', 'monitor',
            '-s', 'before-verification',
            '--need-password-change', 'default',
            'testaccount3@lablup.com',
            '1q2w3e4r',
        ]
        with closing(run(add_arguments)) as p:
            p.expect(EOF)
            assert 'User testaccount3@lablup.com is created' in p.before.decode(), 'Account add error'

    # Check if user is added
    with closing(run(['--output=json', 'admin', 'user', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        user_list = loaded.get('items')

    assert isinstance(user_list, list), 'Expected user list'
    added_user1 = get_user_from_list(user_list, 'testaccount1')
    added_user2 = get_user_from_list(user_list, 'testaccount2')
    added_user3 = get_user_from_list(user_list, 'testaccount3')

    assert bool(added_user1), 'Added account doesn\'t exist: Account#1'
    assert added_user1.get('email') == 'testaccount1@lablup.com', 'E-mail mismatch: Account#1'
    assert added_user1.get('full_name') == 'John Doe', 'Full name mismatch: Account#1'
    assert added_user1.get('status') == 'active', 'User status mismatch: Account#1'
    assert added_user1.get('role') == 'user', 'Role mismatch: Account#1'
    assert added_user1.get('need_password_change') is True, 'Password change status mismatch: Account#1'

    assert bool(added_user2), 'Added account doesn\'t exist: Account#2'
    assert added_user2.get('email') == 'testaccount2@lablup.com', 'E-mail mismatch: Account#2'
    assert added_user2.get('full_name') == 'John Roe', 'Full name mismatch: Account#2'
    assert added_user2.get('status') == 'inactive', 'User status mismatch: Account#2'
    assert added_user2.get('role') == 'admin', 'Role mismatch: Account#2'
    assert added_user2.get('need_password_change') is False, 'Password change status mismatch: Account#2'

    assert bool(added_user3), 'Added account doesn\'t exist: Account#3'
    assert added_user3.get('email') == 'testaccount3@lablup.com', 'E-mail mismatch: Account#3'
    assert added_user3.get('full_name') == 'Richard Roe', 'Full name mismatch: Account#3'
    assert added_user3.get('status') == 'before-verification', 'User status mismatch: Account#3'
    assert added_user3.get('role') == 'monitor', 'Role mismatch: Account#3'
    assert added_user3.get('need_password_change') is True, 'Password change status mismatch: Account#3'


def test_update_user(run: ClientRunnerFunc):
    """
    Run this testcase after test_update_user.
    Testcase for user update.
    TODO: User update with roles is not fully covered yet.
    """
    print("[ Update user ]")

    # Check if user exists
    with closing(run(['--output=json', 'admin', 'user', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        user_list = loaded.get('items')
        assert isinstance(user_list, list), 'Expected user list'

    # Update user
    update_arguments = ['--output=json', 'admin', 'user', 'update', '-u', 'testaccount123', '-n', 'Foo Bar', '-s',
                        'inactive', '-d', 'default', 'testaccount1@lablup.com']
    with closing(run(update_arguments)) as p:
        p.expect(EOF)

    update_arguments = ['--output=json', 'admin', 'user', 'update', '-u', 'testaccount231', '-n', 'Baz Quz', '-s',
                        'active', '-r', 'admin', '--need-password-change', 'testaccount2@lablup.com']
    with closing(run(update_arguments)) as p:
        p.expect(EOF)

    update_arguments = ['--output=json', 'admin', 'user', 'update', '-u', 'testaccount312', '-n', 'Alice B.', '-s',
                        'active', '-r', 'monitor', 'testaccount3@lablup.com']
    with closing(run(update_arguments)) as p:
        p.expect(EOF)

    # Check if user is updated correctly
    with closing(run(['--output=json', 'admin', 'user', 'list'])) as p:
        p.expect(EOF)
        after_update_decoded = p.before.decode()
        after_update_loaded = json.loads(after_update_decoded)
        updated_user_list = after_update_loaded.get('items')
        assert isinstance(updated_user_list, list), 'Expected user list'

    test_user1 = get_user_from_list(updated_user_list, 'testaccount123')
    test_user2 = get_user_from_list(updated_user_list, 'testaccount231')
    test_user3 = get_user_from_list(updated_user_list, 'testaccount312')

    assert bool(test_user1), 'Account not found - Account#1'
    assert test_user1.get('full_name') == 'Foo Bar', 'Full name mismatch: Account#1'
    assert test_user1.get('status') == 'inactive', 'User status mismatch: Account#1'
    assert test_user1.get('role') == 'user', 'Role mismatch: Account#1'
    assert test_user1.get('need_password_change') is False, 'Password change status mismatch: Account#1'
    assert test_user1.get('domain_name') == 'default', 'Domain mismatch: Account#1'

    assert bool(test_user2), 'Account not found - Account#2'
    assert test_user2.get('full_name') == 'Baz Quz', 'Full name mismatch: Account#2'
    assert test_user2.get('status') == 'active', 'User status mismatch: Account#2'
    assert test_user2.get('role') == 'admin', 'Role mismatch: Account#2'
    assert test_user2.get('need_password_change') is True, 'Password change status mismatch: Account#2'

    assert bool(test_user3), 'Account not found - Account#3'
    assert test_user3.get('full_name') == 'Alice B.', 'Full name mismatch: Account#3'
    assert test_user3.get('status') == 'active', 'User status mismatch: Account#3'
    assert test_user3.get('role') == 'monitor', 'Role mismatch: Account#3'
    assert test_user3.get('need_password_change') is False, 'Password change status mismatch: Account#3'


def test_delete_user(run: ClientRunnerFunc):
    """
    !!Run this testcase after running test_add_user
    Testcase for user deletion.
    """
    print("[ Delete user ]")
    with closing(run(['admin', 'user', 'purge', 'testaccount1@lablup.com'])) as p:
        p.sendline('y')
        p.expect(EOF)
        assert 'User is deleted:' in p.before.decode(), 'Account deletion failed: Account#1'

    with closing(run(['admin', 'user', 'purge', 'testaccount2@lablup.com'])) as p:
        p.sendline('y')
        p.expect(EOF)
        assert 'User is deleted:' in p.before.decode(), 'Account deletion failed: Account#2'

    with closing(run(['admin', 'user', 'purge', 'testaccount3@lablup.com'])) as p:
        p.sendline('y')
        p.expect(EOF)
        assert 'User is deleted:' in p.before.decode(), 'Account deletion failed: Account#3'


def test_list_user(run: ClientRunnerFunc):
    """
    Testcase for user listing.
    """
    with closing(run(['--output=json', 'admin', 'user', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        user_list = loaded.get('items')
        assert isinstance(user_list, list)


def get_user_from_list(users: list, username: str) -> dict:
    for user in users:
        if user.get('username') == username:
            return user
    return {}



import json
from contextlib import closing

from ...utils.cli import EOF, ClientRunnerFunc


def test_create_vfolder(run: ClientRunnerFunc):
    """
    Test create vfolder function.
    This test should be executed first in test_vfolder.py.
    TODO: Unannotate the following code after group deletion issue is resolved.
    """
    # Create group first
    # with closing(run(['admin', 'group', 'add', 'default', 'testgroup'])) as p:
    #     p.expect(EOF)
    #     assert 'Group name testgroup is created in domain default' in p.before.decode(), \
    #         'Test group not created successfully.'
    print("[ Create vfolder ]")
    # Create vfolder
    with closing(run(['vfolder', 'create',  '-p', 'rw', 'test_folder1', 'local:volume1'])) as p:
        p.expect(EOF)
        assert 'Virtual folder "test_folder1" is created' in p.before.decode(), 'Test folder1 not created successfully.'

    with closing(run(['vfolder', 'create', '-p', 'ro', 'test_folder2', 'local:volume1'])) as p:
        p.expect(EOF)
        assert 'Virtual folder "test_folder2" is created' in p.before.decode(), 'Test folder2 not created successfully.'

    # Check if vfolder is created
    with closing(run(['--output=json', 'vfolder', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        folder_list = loaded.get('items')
        assert isinstance(folder_list, list), 'Error in listing test folders!'

    test_folder1 = get_folder_from_list(folder_list, 'test_folder1')
    test_folder2 = get_folder_from_list(folder_list, 'test_folder2')

    assert bool(test_folder1), 'Test folder 1 doesn\'t exist!'
    assert test_folder1.get('permission') == 'rw', 'Test folder 1 permission mismatch.'

    assert bool(test_folder2), 'Test folder 2 doesn\'t exist!'
    assert test_folder2.get('permission') == 'ro', 'Test folder 2 permission mismatch.'


def test_rename_vfolder(run: ClientRunnerFunc):
    """
    Test rename vfolder function.
    !! Make sure you execute this test after test_create_vfolder !!
    Otherwise, it will raise an error.
    """
    print("[ Rename vfolder ]")
    # Rename vfolder
    with closing(run(['vfolder', 'rename', 'test_folder1', 'test_folder3'])) as p:
        p.expect(EOF)
        assert 'Renamed' in p.before.decode(), 'Test folder1 not renamed successfully.'

    # Check if vfolder is updated
    with closing(run(['--output=json', 'vfolder', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        folder_list = loaded.get('items')
        assert isinstance(folder_list, list), 'Error in listing test folders!'

    test_folder3 = get_folder_from_list(folder_list, 'test_folder3')
    assert bool(test_folder3), 'Test folder 3 doesn\'t exist!'


def test_delete_vfolder(run: ClientRunnerFunc):
    """
    Test delete vfolder function.
    !! Make sure you execute this test after 1. test_create_vfolder, 2. test_rename_vfolder !!
    Otherwise, it will raise an error.
    """
    print("[ Delete vfolder ]")
    with closing(run(['vfolder', 'delete', 'test_folder2'])) as p:
        p.expect(EOF)
        assert 'Deleted' in p.before.decode(), 'Test folder 2 not deleted successfully.'

    with closing(run(['vfolder', 'delete', 'test_folder3'])) as p:
        p.expect(EOF)
        assert 'Deleted' in p.before.decode(), 'Test folder 3 not deleted successfully.'


def test_list_vfolder(run: ClientRunnerFunc):
    """
    Test list vfolder function.
    """
    with closing(run(['--output=json', 'vfolder', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        folder_list = loaded.get('items')
        assert isinstance(folder_list, list)


def get_folder_from_list(folders: list, foldername: str) -> dict:
    for folder in folders:
        if folder.get('name', '') == foldername:
            return folder
    return {}



import json
from contextlib import closing

from ...utils.cli import EOF, ClientRunnerFunc


def test_add_scaling_group(run: ClientRunnerFunc):
    # Create scaling group
    with closing(run([
        'admin', 'scaling-group', 'add',
        '-d', 'Test scaling group',
        '-i',
        '--driver', 'static',
        '--driver-opts', '{"x": 1}',
        '--scheduler', 'fifo',
        'test_group1',
    ])) as p:
        p.expect(EOF)
        assert 'Scaling group name test_group1 is created.' in p.before.decode(), \
            'Test scaling group not created successfully'

    # Check if scaling group is created
    with closing(run(['--output=json', 'admin', 'scaling-group', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        scaling_group_list = loaded.get('items')
        assert isinstance(scaling_group_list, list), 'Scaling group list not printed properly'

    test_group = get_scaling_group_from_list(scaling_group_list, 'test_group1')
    assert bool(test_group), 'Test scaling group doesn\'t exist'

    # Get the full detail.
    with closing(run(['--output=json', 'admin', 'scaling-group', 'info', 'test_group1'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        scaling_group_list = loaded.get('items')
        assert isinstance(scaling_group_list, list), 'Scaling group info not printed properly'

    test_group = get_scaling_group_from_list(scaling_group_list, 'test_group1')
    assert test_group.get('description') == 'Test scaling group', 'Scaling group description mismatch'
    assert test_group.get('is_active') is False, 'Scaling group active status mismatch'
    assert test_group.get('driver') == 'static', 'Scaling group driver mismatch'
    assert test_group.get('driver_opts') == {'x': 1}, 'Scaling group driver options mismatch'
    assert test_group.get('scheduler') == 'fifo', 'Scaling group scheduler mismatch'
    assert test_group.get('scheduler_opts') == {}, 'Scaling group scheduler options mismatch'


def test_update_scaling_group(run: ClientRunnerFunc):
    # Update scaling group
    with closing(run([
        'admin', 'scaling-group', 'update',
        '-d', 'Test scaling group updated',
        '--driver', 'non-static',
        '--scheduler', 'lifo',
        'test_group1',
    ])) as p:
        p.expect(EOF)
        assert 'Scaling group test_group1 is updated.' in p.before.decode(), \
            'Test scaling group not updated successfully'

    # Check if scaling group is updated
    with closing(run(['--output=json', 'admin', 'scaling-group', 'info', 'test_group1'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        scaling_group_list = loaded.get('items')
        assert isinstance(scaling_group_list, list), 'Scaling group list not printed properly'

    test_group = get_scaling_group_from_list(scaling_group_list, 'test_group1')

    assert bool(test_group), 'Test scaling group doesn\'t exist'
    assert test_group.get('description') == 'Test scaling group updated', 'Scaling group description mismatch'
    assert test_group.get('is_active') is True, 'Scaling group active status mismatch'
    assert test_group.get('driver') == 'non-static', 'Scaling group driver mismatch'
    assert test_group.get('driver_opts') == {'x': 1}, 'Scaling group driver options mismatch'
    assert test_group.get('scheduler') == 'lifo', 'Scaling group scheduler mismatch'
    assert test_group.get('scheduler_opts') == {}, 'Scaling group scheduler options mismatch'


def test_delete_scaling_group(run: ClientRunnerFunc):
    with closing(run(['admin', 'scaling-group', 'delete', 'test_group1'])) as p:
        p.expect(EOF)
        assert 'Scaling group is deleted: test_group1.' in p.before.decode(), 'Test scaling group deletion unsuccessful'


def test_list_scaling_group(run: ClientRunnerFunc):
    with closing(run(['--output=json', 'admin', 'scaling-group', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        scaling_group_list = loaded.get('items')
        assert isinstance(scaling_group_list, list), 'Scaling group list not printed properly'


def get_scaling_group_from_list(scaling_groups: list, groupname: str) -> dict:
    for sg in scaling_groups:
        if sg.get('name') == groupname:
            return sg
    return {}



import json
from contextlib import closing

from ...utils.cli import EOF, ClientRunnerFunc


def test_add_keypair(run: ClientRunnerFunc):
    """
    Test add keypair.
    This test should be execued first in test_keypair.py.
    """
    print("[ Add keypair ]")
    # Add test user
    add_arguments = ['--output=json', 'admin', 'user', 'add', '-u', 'adminkeypair', '-n', 'John Doe',
                     '-r', 'admin', 'default', 'adminkeypair@lablup.com', '1q2w3e4r']
    with closing(run(add_arguments)) as p:
        p.expect(EOF)
        assert 'User adminkeypair@lablup.com is created' in p.before.decode(), 'Account add error'

    add_arguments = ['--output=json', 'admin', 'user', 'add', '-u', 'userkeypair', '-n', 'Richard Doe',
                     'default', 'userkeypair@lablup.com', '1q2w3e4r']
    with closing(run(add_arguments)) as p:
        p.expect(EOF)
        assert 'User userkeypair@lablup.com is created' in p.before.decode(), 'Account add error'

    # Create keypair
    with closing(run(['admin', 'keypair', 'add', '-a', '-i', '-r', '25000', 'adminkeypair@lablup.com', 'default'])) as p:
        p.expect(EOF)
        assert 'Access Key:' in p.before.decode() and 'Secret Key:' in p.before.decode(), 'Keypair add error'

    with closing(run(['admin', 'keypair', 'add', 'userkeypair@lablup.com', 'default'])) as p:
        p.expect(EOF)
        assert 'Access Key:' in p.before.decode() and 'Secret Key:' in p.before.decode(), 'Keypair add error'

    # Check if keypair is added
    with closing(run(['--output=json', 'admin', 'keypair', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        keypair_list = loaded.get('items')
        assert isinstance(keypair_list, list), 'List not printed properly!'

    admin_keypair = get_keypair_from_list(keypair_list, 'adminkeypair@lablup.com')
    user_keypair = get_keypair_from_list(keypair_list, 'userkeypair@lablup.com')

    assert 'access_key' in admin_keypair, 'Admin keypair doesn\'t exist'
    assert admin_keypair.get('is_active') is False, 'Admin keypair is_active mismatch'
    assert admin_keypair.get('is_admin') is True, 'Admin keypair is_admin mismatch'
    assert admin_keypair.get('rate_limit') == 25000, 'Admin keypair rate_limit mismatch'
    assert admin_keypair.get('resource_policy') == 'default', 'Admin keypair resource_policy mismatch'

    assert 'access_key' in user_keypair, 'Admin keypair doesn\'t exist'
    assert user_keypair.get('is_active') is True, 'User keypair is_active mismatch'
    assert user_keypair.get('is_admin') is False, 'User keypair is_admin mismatch'
    assert user_keypair.get('rate_limit') == 5000, 'User keypair rate_limit mismatch'
    assert user_keypair.get('resource_policy') == 'default', 'User keypair resource_policy mismatch'


def test_update_keypair(run: ClientRunnerFunc):
    """
    Test update keypair.
    This test must be executed after test_add_keypair.
    """
    print("[ Update keypair ]")
    # Get access key
    with closing(run(['--output=json', 'admin', 'keypair', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        keypair_list = loaded.get('items')
        assert isinstance(keypair_list, list), 'List not printed properly!'

    admin_keypair = get_keypair_from_list(keypair_list, 'adminkeypair@lablup.com')
    user_keypair = get_keypair_from_list(keypair_list, 'userkeypair@lablup.com')
    assert 'access_key' in admin_keypair, 'Admin keypair info doesn\'t exist'
    assert 'access_key' in user_keypair, 'User keypair info doesn\'t exist'

    # Update keypair
    with closing(run([
        'admin', 'keypair', 'update',
        '--is-active', 'TRUE',
        '--is-admin', 'FALSE',
        '-r', '15000',
        admin_keypair['access_key'],
    ])) as p:
        p.expect(EOF)
        assert 'Key pair is updated:' in p.before.decode(), 'Admin keypair update error'

    with closing(run([
        'admin', 'keypair', 'update',
        '--is-active', 'FALSE',
        '--is-admin', 'TRUE',
        '-r', '15000',
        user_keypair['access_key'],
    ])) as p:
        p.expect(EOF)
        assert 'Key pair is updated:' in p.before.decode(), 'User keypair update error'

    # Check if keypair is updated
    with closing(run(['--output=json', 'admin', 'keypair', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        updated_keypair_list = loaded.get('items')
        assert isinstance(updated_keypair_list, list), 'List not printed properly!'

    updated_admin_keypair = get_keypair_from_list(updated_keypair_list, 'adminkeypair@lablup.com')
    updated_user_keypair = get_keypair_from_list(updated_keypair_list, 'userkeypair@lablup.com')

    assert 'access_key' in updated_admin_keypair, 'Admin keypair doesn\'t exist'
    assert updated_admin_keypair.get('is_active') is True, 'Admin keypair is_active mismatch'
    assert updated_admin_keypair.get('is_admin') is False, 'Admin keypair is_admin mismatch'
    assert updated_admin_keypair.get('rate_limit') == 15000, 'Admin keypair rate_limit mismatch'
    assert updated_admin_keypair.get('resource_policy') == 'default', 'Admin keypair resource_policy mismatch'

    assert 'access_key' in updated_user_keypair, 'Admin keypair doesn\'t exist'
    assert updated_user_keypair.get('is_active') is False, 'User keypair is_active mismatch'
    assert updated_user_keypair.get('is_admin') is True, 'User keypair is_admin mismatch'
    assert updated_user_keypair.get('rate_limit') == 15000, 'User keypair rate_limit mismatch'
    assert updated_user_keypair.get('resource_policy') == 'default', 'User keypair resource_policy mismatch'


def test_delete_keypair(run: ClientRunnerFunc):
    """
    Test delete keypair.
    This test must be executed after test_add_keypair.
    """
    print("[ Delete keypair ]")
    # Get access key
    with closing(run(['--output=json', 'admin', 'keypair', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        keypair_list = loaded.get('items')
        assert isinstance(keypair_list, list), 'List not printed properly!'

    admin_keypair = get_keypair_from_list(keypair_list, 'adminkeypair@lablup.com')
    user_keypair = get_keypair_from_list(keypair_list, 'userkeypair@lablup.com')
    assert 'access_key' in admin_keypair, 'Admin keypair info doesn\'t exist'
    assert 'access_key' in user_keypair, 'User keypair info doesn\'t exist'

    # Delete keypair
    with closing(run(['admin', 'keypair', 'delete', admin_keypair['access_key']])) as p:
        p.expect(EOF)
        print(p.before.decode())

    with closing(run(['admin', 'keypair', 'delete', user_keypair['access_key']])) as p:
        p.expect(EOF)
        print(p.before.decode())

    # Delete test user
    with closing(run(['admin', 'user', 'purge', 'adminkeypair@lablup.com'])) as p:
        p.sendline('y')
        p.expect(EOF)
        assert 'User is deleted:' in p.before.decode(), 'Account deletion failed: adminkeypair'

    with closing(run(['admin', 'user', 'purge', 'userkeypair@lablup.com'])) as p:
        p.sendline('y')
        p.expect(EOF)
        assert 'User is deleted:' in p.before.decode(), 'Account deletion failed: userkeypair'


def test_list_keypair(run: ClientRunnerFunc):
    """
    Test list keypair.
    """
    with closing(run(['--output=json', 'admin', 'keypair', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        keypair_list = loaded.get('items')
        assert isinstance(keypair_list, list), 'List not printed properly!'


def get_keypair_from_list(keypairs: list, userid: str) -> dict:
    for keypair in keypairs:
        if keypair.get('user_id', '') == userid:
            return keypair
    return {}



import json
from contextlib import closing

from ...utils.cli import EOF, ClientRunnerFunc


def test_list_storage(run: ClientRunnerFunc):
    """
    Test list storage.
    """
    print("[ List storage ]")
    with closing(run(['--output=json', 'admin', 'storage', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        storage_list = loaded.get('items')
        assert isinstance(storage_list, list), 'Storage list not printed properly'


def test_info_storage(run: ClientRunnerFunc):
    """
    Test storage info.
    """
    print("[ Print storage info ]")
    with closing(run(['--output=json', 'admin', 'storage', 'info', 'local:volume1'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        storage_list = loaded.get('items')
        assert isinstance(storage_list, list), 'Storage info not printed properly'



import json
from contextlib import closing

from ...utils.cli import EOF, ClientRunnerFunc


def test_add_keypair_resource_policy(run: ClientRunnerFunc):
    print("[ Add keypair resource policy ]")

    # Add keypair resource policy
    add_arguments = [
        'admin', 'keypair-resource-policy', 'add',
        '--default-for-unspecified', 'LIMITED',
        '--total-resource-slots', '{}',
        '--max-concurrent-sessions', '20',
        '--max-containers-per-session', '2',
        '--max-vfolder-count', '15',
        '--max-vfolder-size', '0',
        '--allowed-vfolder-hosts', 'local:volume1',
        '--idle-timeout', '1200',
        'test_krp',
    ]
    with closing(run(add_arguments)) as p:
        p.expect(EOF)
        assert 'Keypair resource policy test_krp is created.' in p.before.decode(), \
            'Keypair resource policy creation not successful'

    # Check if keypair resource policy is created
    with closing(run(['--output=json', 'admin', 'keypair-resource-policy', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        krp_list = loaded.get('items')
        assert isinstance(krp_list, list), 'Keypair resource policy list not printed properly'

    test_krp = get_keypair_resource_policy_from_list(krp_list, 'test_krp')

    assert bool(test_krp), 'Test keypair resource policy doesn\'t exist'
    assert test_krp.get('total_resource_slots') == '{}', 'Test keypair resource policy total resource slot mismatch'
    assert test_krp.get('max_concurrent_sessions') == 20, 'Test keypair resource policy max concurrent session mismatch'
    assert test_krp.get('max_vfolder_count') == 15, 'Test keypair resource policy max vfolder count mismatch'
    assert test_krp.get('max_vfolder_size') == '0 Bytes', 'Test keypair resource policy max vfolder size mismatch'
    assert test_krp.get('idle_timeout') == 1200, 'Test keypair resource policy idle timeout mismatch'
    assert test_krp.get('max_containers_per_session') == 2,\
        'Test keypair resouce policy max containers per session mismatch'
    assert test_krp.get('allowed_vfolder_hosts') == ['local:volume1'], \
        'Test keypair resource policy allowed vfolder hosts mismatch'


def test_update_keypair_resource_policy(run: ClientRunnerFunc):
    print("[ Update keypair resource policy ]")

    # Update keypair resource policy
    add_arguments = [
        'admin', 'keypair-resource-policy', 'update',
        '--default-for-unspecified', 'UNLIMITED',
        '--total-resource-slots', '{}',
        '--max-concurrent-sessions', '30',
        '--max-containers-per-session', '1',
        '--max-vfolder-count', '10',
        '--max-vfolder-size', '0',
        '--allowed-vfolder-hosts', 'local:volume2',
        '--idle-timeout', '1800',
        'test_krp',
    ]
    with closing(run(add_arguments)) as p:
        p.expect(EOF)
        assert 'Update succeeded.' in p.before.decode(), 'Keypair resource policy update not successful'

    # Check if keypair resource policy is updated
    with closing(run(['--output=json', 'admin', 'keypair-resource-policy', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        krp_list = loaded.get('items')
        assert isinstance(krp_list, list), 'Keypair resource policy list not printed properly'

    test_krp = get_keypair_resource_policy_from_list(krp_list, 'test_krp')

    assert bool(test_krp), 'Test keypair resource policy doesn\'t exist'
    assert test_krp.get('total_resource_slots') == '{}', 'Test keypair resource policy total resource slot mismatch'
    assert test_krp.get('max_concurrent_sessions') == 30, 'Test keypair resource policy max concurrent session mismatch'
    assert test_krp.get('max_vfolder_count') == 10, 'Test keypair resource policy max vfolder count mismatch'
    assert test_krp.get('max_vfolder_size') == '0 Bytes', 'Test keypair resource policy max vfolder size mismatch'
    assert test_krp.get('idle_timeout') == 1800, 'Test keypair resource policy idle timeout mismatch'
    assert test_krp.get('max_containers_per_session') == 1,\
        'Test keypair resouce policy max containers per session mismatch'
    assert test_krp.get('allowed_vfolder_hosts') == ['local:volume2'], \
        'Test keypair resource policy allowed vfolder hosts mismatch'


def test_delete_keypair_resource_policy(run: ClientRunnerFunc):
    print("[ Delete keypair resource policy ]")

    # Delete keypair resource policy
    with closing(run(['admin', 'keypair-resource-policy', 'delete', 'test_krp'])) as p:
        p.sendline('y')
        p.expect(EOF)
        assert 'Resource policy test_krp is deleted.' in p.before.decode(), 'Keypair resource policy deletion failed'


def test_list_keypair_resource_policy(run: ClientRunnerFunc):
    print("[ List keypair resource policy ]")
    with closing(run(['--output=json', 'admin', 'keypair-resource-policy', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        krp_list = loaded.get('items')
        assert isinstance(krp_list, list), 'Keypair resource policy list not printed properly'


def get_keypair_resource_policy_from_list(krps: list, name: str) -> dict:
    for krp in krps:
        if krp.get('name') == name:
            return krp
    return {}



import json
from contextlib import closing

from ...utils.cli import EOF, ClientRunnerFunc


def test_add_domain(run: ClientRunnerFunc):
    print("[ Add domain ]")

    # Add domain
    add_arguments = [
        'admin', 'domain', 'add',
        '-d', 'Test domain',
        '-i',
        '--total-resource-slots', '{}',
        '--allowed-vfolder-hosts', 'local:volume1',
        '--allowed-docker-registries', 'cr.backend.ai',
        'test',
    ]
    with closing(run(add_arguments)) as p:
        p.expect(EOF)
        assert 'Domain name test is created.' in p.before.decode(), 'Domain creation not successful'

    # Check if domain is added
    with closing(run(['--output=json', 'admin', 'domain', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        domain_list = loaded.get('items')
        assert isinstance(domain_list, list), 'Domain list not printed properly'

    test_domain = get_domain_from_list(domain_list, 'test')

    assert bool(test_domain), 'Test domain doesn\'t exist'
    assert test_domain.get('description') == 'Test domain', 'Domain description mismatch'
    assert test_domain.get('is_active') is False, 'Domain active status mismatch'
    assert test_domain.get('total_resource_slots') == {}, 'Domain total resource slots mismatch'
    assert test_domain.get('allowed_vfolder_hosts') == ['local:volume1'], 'Domain allowed vfolder hosts mismatch'
    assert test_domain.get('allowed_docker_registries') == ['cr.backend.ai'], 'Domain allowed docker registries mismatch'


def test_update_domain(run: ClientRunnerFunc):
    print("[ Update domain ]")

    # Update domain
    add_arguments = [
        'admin', 'domain', 'update',
        '--new-name', 'test123',
        '--description', 'Test domain updated',
        '--is-active', 'TRUE',
        '--total-resource-slots', '{}',
        '--allowed-vfolder-hosts', 'local:volume2',
        '--allowed-docker-registries', 'cr1.backend.ai',
        'test',
    ]
    with closing(run(add_arguments)) as p:
        p.expect(EOF)
        assert 'Domain test is updated.' in p.before.decode(), 'Domain update not successful'

    # Check if domain is updated
    with closing(run(['--output=json', 'admin', 'domain', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        domain_list = loaded.get('items')
        assert isinstance(domain_list, list), 'Domain list not printed properly'

    test_domain = get_domain_from_list(domain_list, 'test123')

    assert bool(test_domain), 'Test domain doesn\'t exist'
    assert test_domain.get('description') == 'Test domain updated', 'Domain description mismatch'
    assert test_domain.get('is_active') is True, 'Domain active status mismatch'
    assert test_domain.get('total_resource_slots') == {}, 'Domain total resource slots mismatch'
    assert test_domain.get('allowed_vfolder_hosts') == ['local:volume2'], 'Domain allowed vfolder hosts mismatch'
    assert test_domain.get('allowed_docker_registries') == ['cr1.backend.ai'], \
        'Domain allowed docker registries mismatch'


def test_delete_domain(run: ClientRunnerFunc):
    print("[ Delete domain ]")

    # Delete domain
    with closing(run(['admin', 'domain', 'purge', 'test123'])) as p:
        p.sendline('y')
        p.expect(EOF)
        assert 'Domain is deleted:' in p.before.decode(), 'Domain deletion failed'


def test_list_domain(run: ClientRunnerFunc):
    with closing(run(['--output=json', 'admin', 'domain', 'list'])) as p:
        p.expect(EOF)
        decoded = p.before.decode()
        loaded = json.loads(decoded)
        domain_list = loaded.get('items')
        assert isinstance(domain_list, list), 'Domain list not printed properly'


def get_domain_from_list(domains: list, name: str) -> dict:
    for domain in domains:
        if domain.get('name') == name:
            return domain
    return {}



from __future__ import annotations

import click
import subprocess
import sys

from .context import CLIContext
from .utils import CommaSeparatedChoice, CustomUsageArgsCommand


@click.group(invoke_without_command=True, context_settings={'help_option_names': ['-h', '--help']})
@click.pass_context
def main(ctx: click.Context) -> None:
    """
    The integration test suite
    """
    ctx.obj = CLIContext()


@main.command(cls=CustomUsageArgsCommand, context_settings={
    'ignore_unknown_options': True,
    'allow_extra_args': True,
    'allow_interspersed_args': True,
}, usage_args='[PKGS] [PYTEST_ARGS]')
@click.argument("pkgs", type=CommaSeparatedChoice([
    'admin', 'user',
]), metavar='PKGS')
@click.pass_context
def run_cli(
    ctx: click.Context,
    pkgs: list[str],
) -> None:
    """A shortcut command to run pytest against a specific set of CLI-based
    integration tests

    It takes one or more test package names in a comma-separated list (PKGS)
    and forwards all other extra arguments and options (PYTEST_ARGS) to
    the underlying pytest command.

    \b
    Available CLI-based integration test package names:
      admin
      user
    """
    pytest_args = ctx.args
    result = subprocess.run([
        sys.executable, '-m', 'pytest',
        '--pyargs',
        *(f'ai.backend.test.cli_integration.{pkg}' for pkg in pkgs),
        *pytest_args,
    ])
    ctx.exit(result.returncode)


if __name__ == '__main__':
    main()
